import typing
from typing import Any, Iterator

from sema4ai.actions import Row, RowValue, Table

if typing.TYPE_CHECKING:
    import pandas as pd


T = typing.TypeVar("T")


class ResultSet:
    """
    The `ResultSet` is the result of a query.

    The preferred approach is to do a query and then use the `ResultSet`
    to create the structured response that is returned to the LLM by using
    `build_list`.

    Example:

        ```python
        from pydantic import BaseModel
        from sema4ai.actions import Response
        from sema4ai.data import query
        from data_sources import MyDataSource

        class User(BaseModel):
            id: int
            name: str
            email: str

        @query
        def query_users(datasource: MyDataSource) -> Response[list[User]]:
            result_set = datasource.query("SELECT * FROM users")
            return Response(result=result_set.build_list(User))
        ```

    If filtering is required, one can either improve the SQL to filter the data
    (preferred) or filter the result set after it's been returned from the data source
    by iterating over the result set and filtering the rows.

    Example:

        ```python
        from pydantic import BaseModel
        from sema4ai.actions import Response
        from sema4ai.data import query
        from data_sources import MyDataSource

        class User(BaseModel):
            id: int
            name: str
            email: str

        @query
        def query_users(datasource: MyDataSource) -> Response[list[User]]:
            result_set = datasource.query("SELECT * FROM users")
            users_list = [User(**row) for row in result_set if "John" in row["name"]]
            return Response(result=users_list)
        ```

    Another alternative if the shape of the return is not known is to use the `Table`
    class to return a generic structured response.

    Example:

        ```python
        from sema4ai.actions import Response, Table
        from sema4ai.data import query
        from data_sources import MyDataSource


        @query
        def query_users(datasource: MyDataSource) -> Response[Table]:
            result_set = datasource.query("SELECT * FROM users")
            return Response(result=result_set.to_table())
        ```

    The class also has other utility methods to convert the result set to a pandas
    DataFrame (for deeper analysis if needed) or a markdown table (for inspecting
    the result set in the console).

    Note: many methods when converting to a new data type may not really copy the
    data, but rather return a new object that references the same data. This means
    that changing the original result set after it's been converted to a new data type
    may also change the new data type and vice-versa.
    """

    def __init__(self, columns: list[str], rows: list[Row]):
        self._columns = columns
        self._data = rows

    def to_dataframe(self) -> "pd.DataFrame":
        """
        Converts the result set to a pandas DataFrame.

        It's usually not recommended to use this method, unless analysis
        from pandas are actually required (for just returning a value to the
        LLM, it's better to use `build_list` with a structure class -- if the
        shape of the return is known or `to_table` for a generic structured response)
        """
        import pandas as pd

        return pd.DataFrame(self._data, columns=self._columns)

    # Alias for backwards compatibility
    as_dataframe = to_dataframe

    def to_markdown(self) -> str:
        """
        Converts the result set to a markdown table.

        Note: in general it's not recommended to return a markdown table to the LLM,
        instead use the `Table` class for a generic structured response
        (as it allows further structured processing).

        Ideally though, it's recommended to create a custom class and use
        `build_list` to build a list of objects from the result set.
        """
        # Note: str(Table) is expected to return a markdown table!
        return str(self.to_table())

    def __len__(self) -> int:
        """
        Returns the number of rows in the result set.
        """
        return len(self._data)

    def __iter__(self) -> Iterator[dict[str, RowValue]]:
        """
        Iterates over the result set as a list of dictionaries.
        """
        return self.iter_as_dicts()

    def to_table(self) -> Table:
        """
        Converts the result set to a Table.

        Note: changing the columns or rows of the result set after it's been
        converted to a Table will affect the Table and vice-versa.
        """
        return Table(columns=self._columns, rows=self._data)

    def iter_as_dicts(self) -> Iterator[dict[str, RowValue]]:
        """
        Iterates over the result set as a list of dictionaries.

        Note: each dictionary is a row in the result set where
        the keys are the column names and the values are the row values.
        """
        for row in self._data:
            yield dict(zip(self._columns, row))

    def to_dict_list(self) -> list[dict[str, RowValue]]:
        """
        Converts the result set to a list of dictionaries

        Note: each dictionary is a row in the result set where
        the keys are the column names and the values are the row values.
        """
        return list(self.iter_as_dicts())

    def iter_as_tuples(self) -> Iterator[tuple[Any, ...]]:
        """
        Iterates over the result set values as a list of tuples.
        """
        for row in self._data:
            yield tuple(row)

    def build_list(self, item_class: type[T]) -> list[T]:
        """
        Builds a list of objects of the given class from the result set.

        Args:
            item_class: The class to build the list of.

        Returns:
            A list of objects of the given class.

        Note: The class must have a constructor that accepts each item in the
        result set as keyword arguments.

        Example:

            ```python
            from pydantic import BaseModel
            from sema4ai.actions import Response
            from sema4ai.data import query
            from data_sources import MyDataSource

            class User(BaseModel):
                id: int
                name: str
                email: str

            @query
            def query_users(datasource: MyDataSource) -> Response[list[User]]:
                result_set = datasource.query("SELECT * FROM users")
                return Response(result=result_set.build_list(User))
            ```
        """
        return [item_class(**row) for row in self]



import os
import sys
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Literal


@lru_cache
def get_default_sema4ai_home_dir() -> Path:
    home_env_var = os.environ.get("SEMA4AI_HOME")
    if home_env_var:
        home = Path(home_env_var)
    else:
        if sys.platform == "win32":
            localappdata = os.environ.get("LOCALAPPDATA")
            if not localappdata:
                raise RuntimeError("Error. LOCALAPPDATA not defined in environment!")
            home = Path(localappdata) / "sema4ai"
        else:
            # Linux/Mac
            home = Path("~/.sema4ai").expanduser()

    return home


_SPCS_TOKEN_FILE_PATH = Path("/snowflake/session/token")
_LOCAL_AUTH_FILE_PATH = get_default_sema4ai_home_dir() / ".sema4ai" / "sf-auth.json"


@dataclass
class LinkingDetails:
    account: str
    user: str
    role: str
    authenticator: Literal["ID_TOKEN", "OAUTH", "SNOWFLAKE_JWT"]
    applicationUrl: str
    privateKeyPath: str


@dataclass
class SnowflakeAuth:
    linkingDetails: LinkingDetails

    @classmethod
    def from_dict(cls, data: dict) -> "SnowflakeAuth":
        return cls(
            linkingDetails=LinkingDetails(**data["linkingDetails"]),
        )


class SnowflakeAuthenticationError(Exception):
    """Raised when there are authentication-related issues with Snowflake connection."""

    pass


def get_snowflake_connection_details(
    role: str | None = None,
    warehouse: str | None = None,
    database: str | None = None,
    schema: str | None = None,
) -> dict:
    """
    Get Snowflake connection details based on the environment.

    This function first checks if running in SPCS by looking for the token file.
    If found, it uses SPCS authentication, otherwise falls back to local config-based authentication.

    Args:
        role: Snowflake role to use. Falls back to env var
        warehouse: Snowflake warehouse to use. Falls back to env var
        database: Snowflake database to use. Falls back to env var
        schema: Snowflake schema to use. Falls back to env var

    Returns:
        dict: Connection credentials for Snowflake containing environment-specific fields:
            For SPCS:
               host: from SNOWFLAKE_HOST env var
               account: from SNOWFLAKE_ACCOUNT env var
               authenticator: "OAUTH"
               token: from SPCS token file
               role, warehouse, database, schema: from args or env vars
               client_session_keep_alive: True
               port: from SNOWFLAKE_PORT env var
               protocol: "https"
            For local machine:
               account: from config
               user: from config
               role: from args or config
               authenticator: from config (ID_TOKEN, OAUTH, or SNOWFLAKE_JWT)
               warehouse, database, schema: from args
               client_session_keep_alive: True
            Plus authentication-specific fields:
               For ID_TOKEN: session_token and auth_class
               For OAUTH: token
               For SNOWFLAKE_JWT: private_key and private_key_password

    Raises:
        SnowflakeAuthenticationError: If required credentials are missing or invalid
    """
    # Check for SPCS environment first
    import json

    if _SPCS_TOKEN_FILE_PATH.exists():
        token = _SPCS_TOKEN_FILE_PATH.read_text().strip()

        host = os.getenv("SNOWFLAKE_HOST")
        account = os.getenv("SNOWFLAKE_ACCOUNT")

        if not host or not account:
            raise SnowflakeAuthenticationError(
                "Required environment variables SNOWFLAKE_HOST and SNOWFLAKE_ACCOUNT must be set"
            )

        return {
            "host": host,
            "account": account,
            "authenticator": "OAUTH",
            "token": token,
            "role": role or os.getenv("SNOWFLAKE_ROLE"),
            "warehouse": warehouse or os.getenv("SNOWFLAKE_WAREHOUSE"),
            "database": database or os.getenv("SNOWFLAKE_DATABASE"),
            "schema": schema or os.getenv("SNOWFLAKE_SCHEMA"),
            "client_session_keep_alive": True,
            "port": os.getenv("SNOWFLAKE_PORT"),
            "protocol": "https",
        }

    # Fall back to local config-based authentication
    if not _LOCAL_AUTH_FILE_PATH.exists():
        raise SnowflakeAuthenticationError(
            "Not linked to Snowflake, missing authentication data."
        )

    try:
        auth_data = json.loads(_LOCAL_AUTH_FILE_PATH.read_text())
        sf_auth = SnowflakeAuth.from_dict(auth_data)
    except Exception as e:
        raise SnowflakeAuthenticationError(
            f"Failed to read authentication config: {str(e)}"
        )

    config = {
        "account": sf_auth.linkingDetails.account,
        "user": sf_auth.linkingDetails.user,
        "role": role or sf_auth.linkingDetails.role,
        "authenticator": sf_auth.linkingDetails.authenticator,
        "warehouse": warehouse,
        "database": database,
        "schema": schema,
        "client_session_keep_alive": True,
    }

    if sf_auth.linkingDetails.authenticator == "SNOWFLAKE_JWT":
        config["private_key_file"] = sf_auth.linkingDetails.privateKeyPath

    else:
        raise SnowflakeAuthenticationError(
            f"Unsupported authenticator: {sf_auth.linkingDetails.authenticator}"
        )

    return config



import typing
from typing import Optional

if typing.TYPE_CHECKING:
    from sema4ai.actions._action_context import DataContext

    from sema4ai.data._data_server_connection import DataServerConnection
    from sema4ai.data._result_set import ResultSet


class ConnectionNotSetupError(Exception):
    """
    Exception raised when the connection to the data server is not setup.
    """


class _ConnectionHolder:
    _connection: Optional["DataServerConnection"] = None

    @classmethod
    def connection(cls) -> "DataServerConnection":
        if _ConnectionHolder._connection is None:
            raise ConnectionNotSetupError(
                "The connection to the data server is not setup."
            )
        return _ConnectionHolder._connection


class DataSource:
    """
    DataSource is a class that is injected automatically by the framework in `@query` or `@predict`
    and can be used to make queries or run SQL commands in the Data Server.

    It's expected to be annotated with a `DataSourceSpec` with information about the data source
    (this information is used so that clients can know which data source is being used so that
    they can properly configure the connection in the Data Server).

    Example:

        Definition (data_sources.py):

        ```python
        from typing import Annotated
        from sema4ai.data import DataSource, DataSourceSpec

        PostgresDataSource = Annotated[DataSource, DataSourceSpec(
            name="my_datasource",
            engine="postgres",
        )]
        ```

        Usage (data_action.py):
        ```python
        from sema4ai.data import DataSource
        from sema4ai.actions import Response, Table
        from data_sources import PostgresDataSource

        @query
        def my_query(datasource: PostgresDataSource) -> Response[Table]:
            result_set = datasource.query("SELECT * FROM `my_datasource`.my_table")
            return Response(result=result_set.to_table())
        ```
    """

    @property
    def datasource_name(self) -> str:
        """
        The name of the Data Source.

        Note: this may be an empty string if the Data Source is not annotated with a `DataSourceSpec`
        or if it's the result of a Union of multiple data sources received in a `@query` or `@predict`
        function.
        """
        raise NotImplementedError()

    def connection(self) -> "DataServerConnection":
        """
        The connection to the data server.
        """
        return _ConnectionHolder.connection()

    @classmethod
    def setup_connection_from_input_json(cls, value: dict):
        """
        Private API to setup the connection from a JSON object.

        Not expected to be called by users. May be removed in the future.
        """
        from sema4ai.data._connection_provider import _ConnectionProviderFromDict

        connection_provider = _ConnectionProviderFromDict(value)
        connection = connection_provider.connection()
        _ConnectionHolder._connection = connection

    @classmethod
    def setup_connection_from_data_context(cls, data_context: "DataContext"):
        """
        Private API to setup the connection from a DataContext.

        Not expected to be called by users. May be removed in the future.
        """
        from sema4ai.data._connection_provider import (
            _ConnectionProviderFromDataContextOrEnvVar,
        )

        connection_provider = _ConnectionProviderFromDataContextOrEnvVar(
            data_context, "data-server"
        )
        connection = connection_provider.connection()
        _ConnectionHolder._connection = connection

    @classmethod
    def setup_connection_from_env_vars(cls):
        """
        Private API to setup the connection from environment variables.

        Not expected to be called by users. May be removed in the future.
        """
        from sema4ai.data._connection_provider import (
            _ConnectionProviderFromDataContextOrEnvVar,
        )

        connection_provider = _ConnectionProviderFromDataContextOrEnvVar(None, "")
        connection = connection_provider.connection()
        _ConnectionHolder._connection = connection

    @classmethod
    def model_validate(cls, *, datasource_name: str) -> "DataSource":
        """
        Creates a DataSource given its name.

        Return: A DataSource instance with the given value.

        Note: the model_validate method is used for compatibility with
            the pydantic API.
        """

        return _DataSourceImpl(datasource_name)

    def query(
        self,
        query: str,
        params: Optional[list[str | int | float] | dict[str, str | int | float]] = None,
    ) -> "ResultSet":
        """
        API to execute a query and get the result set.

        Args:
            query: The SQL to execute.
            params: The parameters to pass to the query.

                If a list is provided, the parameters are used as positional parameters
                and `?` is used as placeholder.

                If a dict is provided, the parameters are used as named parameters
                and names such as `$key` are used as placeholders.

                Note: the parameters can only be used in the SQL for the replacement of regular variables,
                not for the datasource name (i.e.: it works when the escaping single or double quotes,
                but for datasource names must be escaped with backticks).

                Note: It's expected that the SQL is always using the full data source name (i.e.: `datasource_name.table_name`)
                and not just the table name (i.e.: `table_name`).

        Returns:
            The result set from the query.

        Example:
            result_set = datasource.query("SELECT * FROM `datasource_name`.my_table WHERE id = ?", [1])
            return Response(result=result_set.to_table())

            result_set = datasource.query("SELECT * FROM `datasource_name`.my_table WHERE id = $id", {"id": 1})
            return Response(result=result_set.to_table())
        """
        return self.connection().query(query, params)

    def predict(
        self,
        query: str,
        params: Optional[list[str | int | float] | dict[str, str | int | float]] = None,
    ) -> "ResultSet":
        """
        API to execute a prediction and get the result set.

        Args:
            query: The SQL to execute.
            params: The parameters to pass to the query.

                If a list is provided, the parameters are used as positional parameters
                and `?` is used as placeholder.

                If a dict is provided, the parameters are used as named parameters
                and names such as `$key` are used as placeholders.

                Note: the parameters can only be used in the SQL for the replacement of regular variables,
                not for the datasource name (i.e.: it works when the escaping single or double quotes,
                but for datasource names must be escaped with backticks).

                Note: It's expected that the SQL is always using the full data source name (i.e.: `datasource_name.table_name`)
                and not just the table name (i.e.: `table_name`).

        Returns:
            The result set from the prediction.

        Example:
            result_set = datasource.predict("SELECT * FROM `datasource_name`.my_table WHERE id = ?", [1])
            return Response(result=result_set.to_table())

            result_set = datasource.predict("SELECT * FROM `datasource_name`.my_table WHERE id = $id", {"id": 1})
            return Response(result=result_set.to_table())

        Note: the only difference between `query()` and `predict()` is that `predict()` is expected to be used when
        querying a model.
        """
        return self.connection().predict(query, params)

    def execute_sql(
        self,
        sql: str,
        params: Optional[list[str | int | float] | dict[str, str | int | float]] = None,
    ) -> "ResultSet | None":
        """
        API to execute some SQL. Either it can return a result set (if it was a query) or None
        (if it was some other SQL command).

        Args:
            query: The SQL to execute.
            params: The parameters to pass to the query.

                If a list is provided, the parameters are used as positional parameters
                and `?` is used as placeholder.

                If a dict is provided, the parameters are used as named parameters
                and names such as `$key` are used as placeholders.

                Note: the parameters can only be used in the SQL for the replacement of regular variables,
                not for the datasource name (i.e.: it works when the escaping single or double quotes,
                but for datasource names must be escaped with backticks).

                Note: It's expected that the SQL is always using the full data source name (i.e.: `datasource_name.table_name`)
                and not just the table name (i.e.: `table_name`).

        Returns:
            The result set if the SQL was a query or None if it was some other SQL command.

        Example:
            datasource.run_sql("UPDATE `datasource_name`.my_table SET name = 'John' WHERE id = 1")
        """
        return self.connection().execute_sql(sql, params)


class _DataSourceImpl(DataSource):
    """
    Actual implementation of DataSource (not exposed as we can tweak as needed, only the public API should be relied upon).
    """

    def __init__(self, datasource_name: str):
        self._datasource_name = datasource_name

    @property
    def datasource_name(self) -> str:
        return self._datasource_name



import json

from pydantic import BaseModel, Field
from pydantic.functional_validators import BeforeValidator
from typing_extensions import Annotated


def parse_json_string(v) -> dict | None:
    if isinstance(v, str):
        return json.loads(v)
    return v


class SourceInfo(BaseModel):
    name: Annotated[str, Field(description="The name of the data source", alias="NAME")]
    type: Annotated[str, Field(description="The type of the data source", alias="TYPE")]
    engine: Annotated[
        str | None, Field(description="The engine of the data source", alias="ENGINE")
    ] = None
    connection_data: Annotated[
        dict | None,
        BeforeValidator(parse_json_string),
        Field(
            description="The connection data of the data source",
            alias="CONNECTION_DATA",
        ),
    ] = None


class TableInfo(BaseModel, extra="allow"):
    table_schema: Annotated[
        str, Field(description="The schema of the table", alias="TABLE_SCHEMA")
    ]
    table_name: Annotated[
        str, Field(description="The name of the table", alias="TABLE_NAME")
    ]
    table_type: Annotated[
        str, Field(description="The type of the table", alias="TABLE_TYPE")
    ]
    engine: Annotated[
        str | None, Field(description="The engine of the table", alias="ENGINE")
    ] = None
    version: Annotated[
        int | None, Field(description="The version of the table", alias="VERSION")
    ] = None
    table_rows: Annotated[
        int, Field(description="The estimated number of rows", alias="TABLE_ROWS")
    ]
    create_time: Annotated[
        str, Field(description="The creation timestamp", alias="CREATE_TIME")
    ]
    update_time: Annotated[
        str, Field(description="The last update timestamp", alias="UPDATE_TIME")
    ]


class ColumnInfo(BaseModel):
    column_name: Annotated[
        str, Field(description="The name of the column", alias="COLUMN_NAME")
    ]
    data_type: Annotated[
        str, Field(description="The data type of the column", alias="DATA_TYPE")
    ]


class KnowledgeBaseInfo(BaseModel):
    name: Annotated[str, Field(description="The name of the knowledge base", alias="NAME")]
    project: Annotated[str, Field(description="The project of the knowledge base", alias="PROJECT")]
    model: Annotated[str, Field(description="The model of the knowledge base", alias="MODEL")]
    storage: Annotated[str, Field(description="The storage of the knowledge base", alias="STORAGE")]
    parameters: Annotated[
        dict | None,
        BeforeValidator(parse_json_string),
        Field(
            description="The knowledge base parameters",
            alias="PARAMS",
        ),
    ] = None



"""
The code below could be better. It's a bit too over engineered for what it does
-- it's this way mostly due to historical reasons as initially it was per DataSource and
it was then extracted, so, it could be simplified in the future....
"""
import os
import typing
from typing import Optional, TypedDict

if typing.TYPE_CHECKING:
    from sema4ai.actions._action_context import DataContext

    from sema4ai.data._data_server_connection import DataServerConnection


class _BaseInternalConnectionProvider:
    _connection: Optional["DataServerConnection"] = None

    @property
    def _http_url(self) -> str:
        raise NotImplementedError()

    @property
    def _http_user(self) -> Optional[str]:
        raise NotImplementedError()

    @property
    def _http_password(self) -> Optional[str]:
        raise NotImplementedError()

    @property
    def _mysql_host(self) -> str:
        raise NotImplementedError()

    @property
    def _mysql_port(self) -> int:
        raise NotImplementedError()

    @property
    def _mysql_user(self) -> Optional[str]:
        raise NotImplementedError()

    @property
    def _mysql_password(self) -> Optional[str]:
        raise NotImplementedError()

    @property
    def datasource_name(self) -> str:
        raise NotImplementedError()

    def connection(self) -> "DataServerConnection":
        if self._connection is None:
            from sema4ai.data._data_server_connection import DataServerConnection

            self._connection = DataServerConnection(
                self._http_url,
                self._http_user,
                self._http_password,
                self._mysql_host,
                self._mysql_port,
                self._mysql_user,
                self._mysql_password,
            )

        return self._connection


class HttpDict(TypedDict):
    url: str
    user: str
    password: str


class MysqlDict(TypedDict):
    host: str
    port: int
    user: str
    password: str


class _EnvVarInfoTypeDict(TypedDict):
    http: HttpDict
    mysql: MysqlDict


class _ConnectionProviderFromDataContextOrEnvVar(_BaseInternalConnectionProvider):
    """
    Internal API to wrap a datasource which is passed encrypted.
    """

    _env_var_info: Optional[_EnvVarInfoTypeDict] = None

    def __init__(self, data_context: Optional["DataContext"], path: str):
        """
        Args:
            data_context: The data context.
            path: the path of the data required inside of the data context
                (a '/' splitted path, for instance: 'datasources/my_datasource')
        """
        import json

        self._data_context = data_context
        self._paths = path.split("/")
        env_var_info = os.environ.get("SEMA4AI_DATA_SERVER_INFO")
        if env_var_info:
            self._env_var_info = json.loads(env_var_info)
        else:
            self._env_var_info = None

    def _get_dict_in_data_context(self) -> dict:
        if self._data_context is None:
            return {}

        from robocorp import log

        with log.suppress():
            if self._data_context is None:
                raise RuntimeError("Data context is not available.")

            dct = self._data_context.value

            v = None
            for path in self._paths:
                if not isinstance(dct, dict):
                    dct = None  # Remove from context
                    raise RuntimeError(
                        f"Unable to get path: {self._paths} in data context (expected dict to get {path!r} from)."
                    )
                try:
                    dct = v = dct[path]
                except KeyError:
                    dct = None  # Remove from context
                    raise RuntimeError(
                        f"Unable to get path: {self._paths} in data context (current path: {path!r})."
                    )

            dct = None  # Remove from context
            if v is None:
                raise RuntimeError(
                    f"Error. Path ({self._paths}) invalid for the data context."
                )

            if not isinstance(v, dict):
                del v
                raise RuntimeError(
                    f"Error. Path ({self._paths}) did not map to a dict in the data context."
                )

            return v

    def _get_env_var_info(self) -> Optional[_EnvVarInfoTypeDict]:
        if self._env_var_info is not None:
            return self._env_var_info

        import json

        info = os.environ.get("SEMA4AI_DATA_SERVER_INFO")
        if info:
            self._env_var_info = json.loads(info)
            return self._env_var_info

        return None

    def _get_from_dict(self, path: str) -> typing.Any:
        from robocorp import log

        with log.suppress():
            paths = path.split("/")
            try:
                found = self._get_dict_in_data_context()
                for path in paths:
                    found = found[path]
                return found
            except KeyError:
                info = os.environ.get("SEMA4AI_DATA_SERVER_INFO")
                if info:
                    env_var_info = self._get_env_var_info()
                    if env_var_info:
                        found = typing.cast(dict, env_var_info)
                        try:
                            for path in paths:
                                found = found[path]
                            return found
                        except Exception:
                            pass
                raise RuntimeError(
                    f"Error. Unable to get: {path} (not available in SEMA4AI_DATA_SERVER_INFO env var nor in data context)."
                )

    @property
    def _http_url(self) -> str:
        """
        Provides the actual url wrapped in this class.
        """
        from robocorp import log

        with log.suppress():
            return self._get_from_dict("http/url")

    @property
    def _http_user(self) -> Optional[str]:
        """
        Provides the actual user wrapped in this class.
        """
        from robocorp import log

        with log.suppress():
            try:
                return self._get_from_dict("http/user")
            except Exception:
                return None  # Not mandatory

    @property
    def _http_password(self) -> Optional[str]:
        """
        Provides the actual password wrapped in this class.
        """
        from robocorp import log

        with log.suppress():
            try:
                return self._get_from_dict("http/password")
            except Exception:
                return None  # Not mandatory

    @property
    def _mysql_host(self) -> str:
        from robocorp import log

        with log.suppress():
            return self._get_from_dict("mysql/host")

    @property
    def _mysql_port(self) -> int:
        from robocorp import log

        with log.suppress():
            return self._get_from_dict("mysql/port")

    @property
    def _mysql_user(self) -> Optional[str]:
        from robocorp import log

        with log.suppress():
            return self._get_from_dict("mysql/user")

    @property
    def _mysql_password(self) -> Optional[str]:
        from robocorp import log

        with log.suppress():
            return self._get_from_dict("mysql/password")


class _ConnectionProviderFromDict(_BaseInternalConnectionProvider):
    """
    Internal API to wrap a datasource which is not passed encrypted.
    """

    def __init__(self, value: dict):
        """
        Args:
            value: A dict with the values meant to be wrapped in this class.
        """

        http = value.get("http")
        mysql = value.get("mysql")

        if http is not None:
            self.__http_url = http.get("url")
            self.__http_user = http.get("user")
            self.__http_password = http.get("password")

        if mysql is not None:
            self.__mysql_host = mysql.get("host")
            self.__mysql_port = mysql.get("port")
            self.__mysql_user = mysql.get("user")
            self.__mysql_password = mysql.get("password")

        if self.__http_password is not None:
            from robocorp import log

            log.hide_from_output(self.__http_password)
            log.hide_from_output(repr(self.__http_password))

        if self.__mysql_password is not None:
            from robocorp import log

            log.hide_from_output(self.__mysql_password)
            log.hide_from_output(repr(self.__mysql_password))

    @property
    def _http_url(self) -> str:
        return self.__http_url

    @property
    def _http_user(self) -> Optional[str]:
        return self.__http_user

    @property
    def _http_password(self) -> Optional[str]:
        return self.__http_password

    @property
    def _mysql_host(self) -> str:
        return self.__mysql_host

    @property
    def _mysql_port(self) -> int:
        return self.__mysql_port

    @property
    def _mysql_user(self) -> Optional[str]:
        return self.__mysql_user

    @property
    def _mysql_password(self) -> Optional[str]:
        return self.__mysql_password
